/* icmplog.c of iplog v1.7 -- eric@ojnk.org */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <arpa/inet.h>
#include <time.h>
#include <string.h>
#include <netdb.h>
#define __BSD_SOURCE
#define __FAVOR_BSD
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include "config.h"
#include "common.h"
#include "icmplog.h"
#include "icmp-compat.h"

int main(void)
{
	int sock;
	struct packet pkt;

#ifdef DETECT_SMURF
	signal(SIGALRM, sigalrm);
#endif

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("icmplog", LOGGING_PRIORITY, LOGGING_FACILITY);
	signal(SIGHUP, SIG_IGN);

	fork_to_background();
	initialize_lockfile("icmplog");
	drop_privileges();

	for (;;) {
#ifdef DETECT_SMURF
		errno = 0;
		read(sock, (struct icmp_packet *)&pkt, sizeof(pkt));
		if (errno == EINTR)
			continue;
#else
		read(sock, (struct icmp_packet *)&pkt, sizeof(pkt));
#endif

		if (pkt.ip.ip_hl != 5) {
			syslog(LEVEL, "suspicious ip options from %s", inet_ntoa(src));
			continue;
		}
		switch (MIN(pkt.icmp.icmp_type, 19)) {
			case ICMP_TIME_EXCEEDED:
				break;			/* ignored */
			case ICMP_ECHO_REPLY:
#ifdef DETECT_SMURF
				if (smurf((struct gen_packet *)&pkt))
					syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), UNFIX(pkt.ip.ip_len) - ICMP_SIZE, ICMP_SIZE);
				break;
#endif
			case ICMP_ECHO:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), UNFIX(pkt.ip.ip_len) - ICMP_SIZE, ICMP_SIZE);
				break;
			case ICMP_DEST_UNREACHABLE:
				if (pkt.icmp.icmp_code > UNREACHABLE_MAX) {
#ifdef LOG_WEIRD_ICMP
					syslog(LEVEL, "undefined destination unreachable code from %s(?)", inet_ntoa(src));
#endif
					break;
				}
				syslog(LEVEL, unreach[(int)pkt.icmp.icmp_code], inet_ntoa(src), protolookup(proto), UNFIX(pkt.tcp.th_dport), UNFIX(pkt.tcp.th_sport));
				break;
			case ICMP_REDIRECT:
				if (pkt.icmp.icmp_code > REDIRECT_MAX) {
#ifdef LOG_WEIRD_ICMP
					syslog(LEVEL, "undefined redirect code from %s(?)", inet_ntoa(src));
#endif
					break;
				}
				print_redirect(pkt);
				break;
			case ICMP_TIMESTAMP_REPLY:
				print_timestamp_reply(pkt);
				break;
			case ICMP_ADDRESS_REPLY:
				print_address_reply(pkt);
				break;
			case ICMP_PARAMETER_PROBLEM:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), pkt.icmp.icmp_pptr);
				break;
			case ICMP_ROUTER_ADVERT:
				print_router_advertisement(pkt);
				break;
			case ICMP_UNDEFINED:
#ifdef LOG_WEIRD_ICMP
				syslog(LEVEL, "undefined icmp type %d from %s(?)", (int)pkt.icmp.icmp_type, inet_ntoa(src));
#endif
				break;
			default:
				if (codes[(int)pkt.icmp.icmp_type])
					syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src));
#ifdef LOG_WEIRD_ICMP
				else
					syslog(LEVEL, "undefined icmp type %d from %s(?)", (int)pkt.icmp.icmp_type, inet_ntoa(src));
#endif
				break;
		}
	}
}

static char *protolookup(int num)
{
	struct protoent *pe;

	if ((pe = getprotobynumber(num)) != NULL)
		return (pe->p_name);
	else {
		static char buff[16];

		sprintf(buff, "proto %d", num);
		return (buff);
	}
}

static void print_redirect(struct packet pkt)
{
	char gate[18], saddr[18], daddr[18];

	strcpy(gate, inet_ntoa(gw));
	strcpy(saddr, inet_ntoa(src));
	strcpy(daddr, inet_ntoa(dest));

	syslog(LEVEL, redir[(int)pkt.icmp.icmp_code], gate, saddr, daddr);

	return;
}

static void print_address_reply(struct packet pkt)
{
	char saddr[18], daddr[18];
	struct in_addr in;

	in.s_addr = htonl(pkt.icmp.icmp_mask);
	strcpy(saddr, inet_ntoa(src));
	strcpy(daddr, inet_ntoa(in));

	syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], saddr, daddr);

	return;
}


static void print_router_advertisement(struct packet pkt)
{
	struct id_rdiscovery *routers;
	char msg[768], temp[64], duration[32];
	u_char n;
	u_int time = UNFIX(pkt.icmp.icmp_lifetime);

	memset(msg, 0, sizeof(msg));
	routers = (struct id_rdiscovery *)&pkt.icmp.icmp_dun.id_rdiscovery;

	for (n = 1; n <= MIN(pkt.icmp.icmp_num_addr, 12); n++, routers++) {
		memset(temp, 0, sizeof(temp));
		snprintf(temp, 64, "%s%s", (n > 1) ? " -> " : "", inet_ntoa(routers->router_addr));
		snprintf(msg, 768, "%s%s", msg, temp);
	}

	if (time < 60)
		sprintf(duration, "%u sec", time);
	else
		sprintf(duration, "%u min, %u sec", (time / 60), (time % 60));

	syslog(LEVEL, "path of router advertisement from %s: %s (lifetime %s)", inet_ntoa(src), msg, duration);

	return;
}

static void print_timestamp_reply(struct packet pkt)
{
	char ret[16];
	struct tm *tm;
	u_long actual_time;

	actual_time = ntohl(pkt.icmp.icmp_ttime) / 1000;
	tm = localtime((time_t *) & actual_time);
	sprintf(ret, "%.2d:%.02d:%.02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
	syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), ret);

	return;
}
