/* tcplog.c of iplog v1.7 -- eric@ojnk.org */
#include <unistd.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#define __BSD_SOURCE
#define __FAVOR_BSD
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <errno.h>
#include <ctype.h>
#include <arpa/inet.h>
#include "config.h"
#include "common.h"
#include "ident.h"
#include "tcplog.h"

#ifdef RETRIEVE_IDENTD
void sigchld(int stat)
{
	while (waitpid(-1, &stat, WNOHANG) > 0);
	signal(SIGCHLD, sigchld);
	return;
}
#endif

#if defined(USE_DNS_CACHE) || defined(USE_PORTSCAN_DETECTION)
void sighup(int stat)
{
	syslog(LEVEL, "caught hangup signal -- clearing cache");
#ifdef USE_DNS_CACHE
	init_dns_cache();
#endif
#ifdef USE_PORTSCAN_DETECTION
	init_portscan();
#endif
	signal(SIGHUP, sighup);
	return;
}

time_t last_alarm = 0;
void sigalrm(int stat)
{
	if ((last_alarm + 10) <= time(NULL)) {
#ifdef USE_DNS_CACHE
		expire_dns();
#endif
#ifdef USE_PORTSCAN_DETECTION
		expire_portscan();
#endif
		time(&last_alarm);
	}
	signal(SIGALRM, sigalrm);
	alarm(15);
}

#endif

int main(void)
{
	int sock;
	time_t tm;
	struct header hdr;

#if defined(DETECT_FIN) || defined(DETECT_NULL)
	unsigned short flags = 0;

#endif
#ifdef DETECT_FIN
	time_t finlast = 0;
	u_char fincount = 0;
	u_long fin_last = 0;

#endif
#ifdef DETECT_NULL
	time_t nulllast = 0;
	u_long null_last = 0;

#endif
#ifdef DETECT_BOGUS
	u_long last_bogus = 0;

#endif
#ifdef RETRIEVE_IDENTD
	char *user = NULL;

#endif
	time_t synlast = 0;
	u_char syncount = 0;

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("tcplog", LOGGING_OPTIONS, LOGGING_FACILITY);
#ifdef RETRIEVE_IDENTD
	signal(SIGCHLD, sigchld);
#endif
#if defined(USE_PORTSCAN_DETECTION) || defined(USE_DNS_CACHE)
	signal(SIGHUP, sighup);
	signal(SIGALRM, sigalrm);
#endif

#ifdef USE_PORTSCAN_DETECTION
	init_portscan();
#endif
#ifdef USE_DNS_CACHE
	init_dns_cache();
#else
	signal(SIGHUP, SIG_IGN);
#endif

	fork_to_background();
	initialize_lockfile("tcplog");
	drop_privileges();

#if defined(USE_PORTSCAN_DETECTION) || defined(USE_DNS_CACHE)
	alarm(15);
	errno = 0;
#endif

	for (;;) {
#if defined(USE_PORTSCAN_DETECTION) || defined(USE_DNS_CACHE)
		errno = 0;
		read(sock, (struct header *)&hdr, sizeof(hdr));
		if (errno == EINTR)
			continue;
#else
		read(sock, (struct header *)&hdr, sizeof(hdr));
#endif
		time(&tm);

		if (syncount) {
			syncount -= ((tm - synlast) >> 1);
			synlast = tm;
		}
#ifdef DETECT_BOGUS
		/* Nmap and Queso use a bogus tcp flag to "fingerprint" OS'es.. */
		if ((hdr.tcp.th_flags & TH_BOG) && last_bogus != hdr.ip.ip_src.s_addr) {
			last_bogus = hdr.ip.ip_src.s_addr;
			syslog(LEVEL, "bogus tcp flags set by %s (%s)", hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), inet_ntoa(hdr.ip.ip_src));
		}
#endif

#ifdef DETECT_NULL
		if (!hdr.tcp.th_flags) {
			if (tm >= (nulllast + 60))
				flags &= ~NULL_LOGGED;
			if ((hdr.ip.ip_src.s_addr == null_last) && !(flags & NULL_LOGGED)) {
				syslog(LEVEL, "null scan detected from %s (%s)", hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), inet_ntoa(hdr.ip.ip_src));
				flags |= NULL_LOGGED;
				nulllast = tm;
			} else if (hdr.ip.ip_src.s_addr != null_last) {
				flags &= ~NULL_LOGGED;
				null_last = hdr.ip.ip_src.s_addr;
			}
			continue;
		}
#endif

#ifdef DETECT_FIN
		if ((hdr.tcp.th_flags & TH_FIN) && !(hdr.tcp.th_flags & TH_ACK)) {
			if (tm >= (finlast + 60))
				flags &= ~FIN_LOGGED;
			if ((hdr.ip.ip_src.s_addr == fin_last) && !(flags & FIN_LOGGED)) {
				if (++fincount >= 2) {
					syslog(LEVEL, "%s scan detected from %s (%s)", (hdr.tcp.th_flags & TH_URG) ? "Xmas" : "fin", hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), inet_ntoa(hdr.ip.ip_src));
					flags |= FIN_LOGGED;
					finlast = tm;
				}
			} else if (hdr.ip.ip_src.s_addr != fin_last) {
				flags &= ~FIN_LOGGED;
				fincount = 1;
				fin_last = hdr.ip.ip_src.s_addr;
			}
			continue;
		}
#endif
		if ((hdr.tcp.th_flags & TH_SYN) && !(hdr.tcp.th_flags & TH_ACK)) {
			synlast = tm;

			if (syncount != SYN_FLOOD)
				syncount++;

#ifdef USE_PORTSCAN_DETECTION
			if (portscan(&hdr.ip.ip_src))
				continue;
#endif

#ifndef LOG_HTTP_CONNECTS
			if (((hdr.tcp.th_dport == UNFIX(80)) || (hdr.tcp.th_dport == UNFIX(8080)) || (hdr.tcp.th_dport == UNFIX(1080))) && UNFIX(hdr.tcp.th_sport) >= 1024)
				continue;
#endif

#ifndef LOG_FTP_TRANSFERS
			if (hdr.tcp.th_sport == UNFIX(20) && UNFIX(hdr.tcp.th_dport) >= 1024)
				continue;
#endif
#ifdef RETRIEVE_IDENTD
			if (hdr.tcp.th_dport != IDENTD_PORT && (hdr.tcp.th_sport != FTP_PORT || UNFIX(hdr.tcp.th_dport) <= 1024) && (syncount != SYN_FLOOD) && l_port(hdr.tcp.th_dport)) {
				if (!fork()) {
					user = get_identd_data(hdr.tcp.th_sport, hdr.tcp.th_dport, hdr.ip.ip_src.s_addr, hdr.ip.ip_dst.s_addr);

					if (user != NULL) {
						syslog(LEVEL, "%s connection attempt from %s@%s:%u", servlookup(hdr.tcp.th_dport), user, hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), UNFIX(hdr.tcp.th_sport));
						free(user);
					} else
						syslog(LEVEL, "%s connection attempt from %s:%u", servlookup(hdr.tcp.th_dport), hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), UNFIX(hdr.tcp.th_sport));

					exit(0);
				}
			} else
				syslog(LEVEL, "%s connection attempt from %s:%u", servlookup(hdr.tcp.th_dport), hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), UNFIX(hdr.tcp.th_sport));
		}
#else
			syslog(LEVEL, "%s connection attempt from %s:%u", servlookup(hdr.tcp.th_dport), hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)), UNFIX(hdr.tcp.th_sport));
		}
#endif
	}
	(void)close(sock);
}

static char *servlookup(u_short port)
{
	struct servent *se;

	if ((se = getservbyport(port, "tcp")) != NULL)
		return (se->s_name);
	else {
		static char buff[16];

		sprintf(buff, "port %u", UNFIX(port));
		return (buff);
	}
}

static char *hostlookup(u_long addr, int dohost)
{
	struct in_addr i;

	i.s_addr = addr;

	if (!dohost) {
		return (inet_ntoa(i));
	} else {
#ifdef USE_DNS_CACHE
		char *host;

		if ((host = get_dns_cache(addr)) != NULL)
			return (host);
		else
			return (inet_ntoa(i));
	}
#else
		struct hostent *he;
		he = gethostbyaddr((char *)&i, sizeof(struct in_addr), AF_INET);

		if (he == NULL)
			return (inet_ntoa(i));
		else
			return (he->h_name);
	}
#endif
	return ("unknown");
}
